/*
** Copyright (C) 2020 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/

#include "HmiPostProcessingManager.h"

#include <cstring>

namespace
{
/// Unaligned size of buffer for lut texture
constexpr int LutSize = LUT_SIZE * LUT_SIZE * 3;

/// Aligned size of buffer for lut texture
constexpr int LutSizeAligned = LUT_SIZE * LUT_SIZE * 4;
}

/******************************************************************************
 * HmiPostProcessorAdapter class defenition
 *****************************************************************************/

HmiPostProcessorAdapter::HmiPostProcessorAdapter(QQuickView *view)
    : m_view(view)
{
    qDebug() << "Init view is null:" << m_view.isNull();
}

void HmiPostProcessorAdapter::sync() const
{
    if (!m_view.isNull())
    {
        m_view->update();
    }
}

void HmiPostProcessorAdapter::bind() const
{
    if (m_view.isNull())
    {
        qCritical() << "Invalid initialization. View isn't available for post processor";
    }
    else
    {
        QObject::connect(m_view, &QQuickView::sceneGraphInitialized,
                         this,  &HmiPostProcessorAdapter::onGLInited,
                         Qt::DirectConnection);
        QObject::connect(m_view, &QQuickView::afterRendering,
                         this,  &HmiPostProcessorAdapter::onGLRendering,
                         Qt::DirectConnection);
    }
}

void HmiPostProcessorAdapter::setColorProfilingLut(const QByteArray &data)
{
    // Print (check) version of color correction profile
    qInfo() << "Color profile version:" << static_cast<uint8_t>(*data.constData());

    // Set lut data for post processor
    m_lut.fill(0, ::LutSizeAligned);
    std::memcpy(m_lut.data(), data.constData() + 1, ::LutSize);

    // Configurate post processor
    m_proc.setColorProfile(reinterpret_cast<const unsigned char*>(m_lut.constData()),
                           static_cast<size_t>(m_lut.size()));
}

QByteArray HmiPostProcessorAdapter::colorProfilingLut() const
{
    return m_lut;
}

bool HmiPostProcessorAdapter::isLutValid() const
{
    return m_lut.size() > 367;
}

void HmiPostProcessorAdapter::setPrewapringState(bool enabled)
{
    m_proc.turnWarpingOn(enabled);
}

void HmiPostProcessorAdapter::setPrewarpingTable(const QByteArray &data, const QSize &originSize,
                                                 const QRect &prewarpingGeometry)
{
    m_proc.setWarpingTables(reinterpret_cast<const unsigned char*>(data.constData()),
                            static_cast<size_t>(data.size()),  DEFAULT_WARPING_TABLE_COLUMNS,
                            DEFAULT_WARPING_TABLE_ROWS, originSize.width(), originSize.height(),
                            prewarpingGeometry.x(), prewarpingGeometry.y(),
                            prewarpingGeometry.width(), prewarpingGeometry.height());
}

void HmiPostProcessorAdapter::setPrewarpingValue(int percent)
{
    m_proc.setWarpingPercentage(qBound(0.0f, 0.01f * percent, 1.0f));
}

void HmiPostProcessorAdapter::onGLInited()
{
    if (!m_view.isNull())
    {
        uint32_t fbo = m_proc.init(static_cast<uint32_t>(m_view->width()),
                                   static_cast<uint32_t>(m_view->height()));

        qDebug() << "fbo ="    << fbo
                 << "width ="  << static_cast<uint32_t>(m_view->width())
                 << "height =" << static_cast<uint32_t>(m_view->height())
                 << "size ="   << m_view->size();

        m_view->setRenderTarget(fbo, m_view->size());
    }
    else
    {
        qWarning() << "GL initialization fialed";
    }
}

void HmiPostProcessorAdapter::onGLRendering()
{
    m_proc.postProcess();
}


/******************************************************************************
 * HmiPostProcessingManager class defenition
 *****************************************************************************/

HmiPostProcessingManager::HmiPostProcessingManager() : m_inversiveLut(true)
{
}

void HmiPostProcessingManager::registrate(HmiScreenId screenId, QQuickView *view)
{
    auto const id = static_cast<int>(screenId);

    if (screenId == HmiScreenId::Undefined)
    {
        qWarning() << "Couldn't register post processor for undefined screen";
        return;
    }

    if (!view)
    {
        qWarning() << "Couldn't register post processor for null view instance";
        return;
    }

    if (m_map.find(id) != m_map.cend())
    {
        qDebug() << "Re-registration of post processor";
        return;
    }

    m_map.emplace(id, ProcPtr(new HmiPostProcessorAdapter(view)));
}

void HmiPostProcessingManager::initialize()
{
    if (m_map.empty())
    {
        qDebug() << "There aren't any registrated postprocessors. "
                    "Skip initialization step";
    }
    else
    {
        loadColorCorrectionProfiles();
        loadPrewarpingProfile();
        subscribePostProcessing();
    }
}

QPair<bool, QByteArray> HmiPostProcessingManager::colorCorrectionData(HmiScreenId screenId)
{
    QList<QString> displayNames = {"Center", "Secondary", "Cluster", "Hud"};
    bool isValid = true;
    QString resultString("Simulated value for " + displayNames[screenId] + " display");
    QByteArray colorProfile(resultString.toStdString().c_str());
    QPair<bool, QByteArray> result(isValid,colorProfile);
    return result;
}

void HmiPostProcessingManager::setColorProfilingLut(HmiScreenId screenId, const QByteArray &colorProfile)
{
    auto &postProcessingAdapter = m_map[screenId];
    if (postProcessingAdapter == nullptr)
    {
        qWarning() << "Display" << screenId << "is not present. Skip request";
        return;
    }
    postProcessingAdapter->setColorProfilingLut(colorProfile);
}

void HmiPostProcessingManager::resetColorProfileParameters()
{
    qInfo() << "reset color profile parameters";
}

void HmiPostProcessingManager::setLutInversion(bool inv)
{
    m_inversiveLut = inv;
}

void HmiPostProcessingManager::loadColorCorrectionProfiles()
{
    qDebug() << "loadColorCorrectionProfiles";
    QByteArray byteArray(::LutSize + 1, 0);
    byteArray[0] = '\x01'; // Version

    auto getLut = [] (QByteArray &arr, bool straight = true) {

        for (uint32_t i = 1, c = straight ? 0 : 255; i < ::LutSize + 1; ++i)
        {
            arr[i] = c;
            c = straight ? (c != 255) * (c + 1) : (c != 1) * (c - 1);
        }
    };

    getLut(byteArray, m_inversiveLut); //TODO run getLut(byteArray, false) to inverse colors

    for (auto &it : m_map)
    {
        auto &postProcessorAdapter = it.second;
        postProcessorAdapter->setColorProfilingLut(byteArray);
    }
}

void HmiPostProcessingManager::loadPrewarpingProfile()
{
    // Nothing to be done here
}

void HmiPostProcessingManager::subscribePostProcessing()
{
    for (auto &it : m_map)
    {
        qDebug() << "Try binding post processing for screen:" << it.first;

        it.second->bind();
        it.second->sync();
    }

    HmiRender::setErrorCallback(&errorHandlerCallback);
}

void HmiPostProcessingManager::errorHandlerCallback(const char *message)
{
    qWarning() << message;
}
