/*
** Copyright (C) 2020 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/

#ifndef HMIPOSTPROCESSINGMANAGER_H
#define HMIPOSTPROCESSINGMANAGER_H

#include "PostProcessor/HmiPostProcessor.h"
#include <QQuickView>
#include <QPointer>

#include <map>
#include <memory>

/**
 * @class HmiPostProcessorAdapter
 * The class associates quick view with
 * post processor and corresponded data.
  */
class HmiPostProcessorAdapter Q_DECL_FINAL : public QObject
{
    Q_OBJECT
    Q_DISABLE_COPY(HmiPostProcessorAdapter)

public:
    /**
     * Default constructor.
     */
    explicit HmiPostProcessorAdapter(QQuickView *view);

    /**
     * Syncs associated view and post processor.
     */
    void sync() const;

    /**
     * Bind post processor with view render pipeline.
     */
    void bind() const;

    /**
     * Sets lut with color correction profile data.
     */
    void setColorProfilingLut(const QByteArray &data);

    /**
     * Gets lut with color correction profile data.
     */
    QByteArray colorProfilingLut() const;

    /**
     * Checks color correction profile data is correct
     */
    bool isLutValid() const;

    /**
     * Activates prewarping in post processor.
     */
    void setPrewapringState(bool enabled = true);

    /**
     * Sets table for prewarping.
     */
    void setPrewarpingTable(const QByteArray &data, const QSize &originSize = QSize(-1, -1),
                            const QRect &prewarpingGeometry = QRect(-1, -1, -1, -1));

    /**
     * Sets value of prewarping in percents
     */
    void setPrewarpingValue(int percent);

public slots:
    /**
     * Inits post processor.
     */
    void onGLInited();

    /**
     * Post-processing execution.
     */
    void onGLRendering();

private:
    QPointer<QQuickView> m_view;
    QByteArray           m_lut;
    HmiPostProcessor     m_proc;
};


/**
 * @class HmiPostProcessingManager
 * This class manages of post processing.
 * It allows to activate color correction and prewarping.
  */
class HmiPostProcessingManager Q_DECL_FINAL
{
    Q_DISABLE_COPY(HmiPostProcessingManager)

public:

    /**
     * @enum PrewarpingState
     */
    enum class PrewarpingState
    {
        Absence = 0, // Prewarping won't be used
        Static,      // Prewarping uses default percent value
        Dynamic      // Prewarping value is taken from DisplayManager
    };

    enum HmiScreenId
    {
        Undefined = -1,
        Central = 0,
        Secondary = 1,
        NavOnCluster = 2,
        Hud = 3 // HUD should be never synchronized with AP
    };

public:
    /** Default constructor */
    explicit HmiPostProcessingManager();

    /**
     * The method registrates post processor for specified view.
     * The post processor will be empty until coressponded data is loaded.
     * @param[in] screenId - id of view for post processing
     * @param[in] view - instance of view for post processing
     */
    void registrate(HmiScreenId screenId, QQuickView *view);

    /**
     * The method loads data for post processing.
     *
     * Loading is consisted from two stages:
     * - Reading configuration for color correction
     * - Reading configuration for prewarping (HUD only)
     */
    void initialize();

    /**
     * the method returns color profile parameters from screen id
    */
    QPair<bool, QByteArray> colorCorrectionData(HmiScreenId screenId);

    /**
     * the method sets color profile parameters for screen Id
    */
    void setColorProfilingLut(HmiScreenId screenId, const QByteArray& colorProfile);

    /**
     * the method resets color profile parameters
    */
    void resetColorProfileParameters();

    void setLutInversion(bool inv);

private:
    /**
     * Loads color correction profiles data for each registered view.
     */
    void loadColorCorrectionProfiles();

    /**
     * Loads prewarping tables data for registered HUD view.
     */
    void loadPrewarpingProfile();

    /**
     * Binds GL pipline view with post processing.
     */
    void subscribePostProcessing();

    /**
     * The method returns preset state of prewarping
     * for HUD display. The state is read from PCL.
     * @see PrewarpingState
     */
    static PrewarpingState prewarpingState();

    /**
     * The method returns preset data of prewarping tables
     * for HUD display. The data is read from PCL.
     */
    static QByteArray prewarpingTableData();

    /**
     * Callback function for detecting occured errors
     * @param[in] error pointer of error data object
     */
    static void errorHandlerCallback(const char *message);

private:
    using ProcPtr = std::unique_ptr<HmiPostProcessorAdapter>;

    /// Auxiliary structure for binding post processing data and processor.
    std::map<int, ProcPtr> m_map;
    bool m_inversiveLut;
};

#endif // HMIPOSTPROCESSINGMANAGER_H
