/*
** Copyright (C) 2019 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/

#include "HmiPostProcessor.h"

#include <sstream>
#include <cstdint>
#include <float.h>

#define WARPING_TABLE_CELL_SIZE 4
#define COLOR_PROFILE_SIZE (LUT_SIZE * LUT_SIZE * LUT_SIZE * 3)
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))

static const char *vertexShader =
#if !defined (INTEGRITY) && !defined (GENIVI)
"#version 330 core\n"
#else
"#version 300 es\n"
#endif
"layout(location = 0) in vec4 a_attr0;\n\
layout(location = 1) in vec4 a_attr1;\n\
uniform float factor;\n\
uniform mat4 mvp;\n\
out vec2 uv;\n\
void main() {\n\
    float min_factor = min(factor / 0.5, 1.0);\n\
    float max_factor = max((factor - 0.5) / 0.5, 0.0);\n\
    uv = a_attr0.zw;\n\
    gl_Position = mvp * vec4(mix(a_attr0.xy, mix(a_attr1.xy, a_attr1.zw, max_factor), min_factor), 0.0, 1.0);\n\
}";

static const char *fragmentShader =
#if !defined (INTEGRITY) && !defined (GENIVI)
"#version 330 core\n"
#else
"#version 300 es\n"
#endif
"uniform sampler2D rgb;\n\
uniform sampler2D lut;\n\
uniform sampler2D mask;\n\
uniform vec4 maskPosition;\n\
uniform vec4 opacity;\n\
in vec2 uv;\n\
out vec4 color;\n\
void main() {\n\
    vec4 origin = texture(rgb, uv);\n\
    vec2 flip = uv * maskPosition.xy + maskPosition.zw;\n\
    flip = vec2(flip.x, ((opacity.w < 0.5) ? (1.0 - flip.y) : flip.y));\n\
    vec4 channels = texture(mask, flip);\n\
    float alpha = 1.0 - ((channels.b * opacity.y) + (channels.g * opacity.z)) * opacity.x;\n\
    color = vec4(texture(lut, vec2(origin.x, 0.0)).x,\n\
                 texture(lut, vec2(origin.y, 0.25)).x,\n\
                 texture(lut, vec2(origin.z, 0.50)).x, origin.w) * alpha;\n\
}";

HmiPostProcessor::HmiPostProcessor() :
    m_isWarpingOn(false),
    m_warpingTableColumnsCount(DEFAULT_WARPING_TABLE_COLUMNS),
    m_warpingTableRowsCount(DEFAULT_WARPING_TABLE_ROWS),
    m_warpingPercentage(0.0f),
    m_Width(0),
    m_Height(0),
    m_OffsetX(0),
    m_OffsetY(0),
    m_colorProfile(nullptr),
    m_maskTexture(0),
    m_defaultFbo(0),
    m_indexCount(0)
{
    turnWarpingOn(false);

    m_maskPosition[0] = 1.0f;
    m_maskPosition[1] = 1.0f;
    m_maskPosition[2] = 0.0f;
    m_maskPosition[3] = 0.0f;

    m_maskOpacity[0] = 0.0f;
    m_maskOpacity[1] = 0.0f;
    m_maskOpacity[2] = 0.0f;
    m_maskOpacity[3] = 0.0f;
}

uint32_t HmiPostProcessor::init(uint32_t width, uint32_t height, uint32_t defaultFbo)
{
    m_defaultFbo = defaultFbo;

    if (!HmiRender::init())
    {
        return 0;
    }

    m_Width = width;
    m_Height = height;

    m_Target.create(width, height, true);
    m_lutTexture.create(LUT_SIZE * LUT_SIZE, 4, nullptr, HmiTexture::R, false, false);

    uint8_t data[4];
    data[0] = 0;
    data[1] = 0;
    data[2] = 0;
    data[3] = 0;
    m_whiteTexture.create(1, 1, data, HmiTexture::RGBA, false, false);

    if (m_Shader.load(vertexShader, fragmentShader))
    {
        m_Shader.setAttributes({"a_attr0", "a_attr1"});
        m_Shader.addTexture("rgb", m_Target.m_color.m_texture);
        m_Shader.addTexture("lut", m_lutTexture.m_texture);
        m_maskTexture = m_Shader.addTexture("mask", m_whiteTexture.m_texture);
        m_Shader.addUniform("factor", &m_warpingPercentage, 1);
        m_Shader.addUniform("opacity", m_maskOpacity, 4);
        m_Shader.addUniform("maskPosition", m_maskPosition, 4);

        m_Shader.m_mvp = m_mvp.mat;

        // Generating flat rectangle
        m_FlatMesh.setData({{0.0f, 0.0f, 0.0f, 0.0f,
                             1.0f, 0.0f, 1.0f, 0.0f,
                             1.0f, 1.0f, 1.0f, 1.0f,
                             0.0f, 1.0f, 0.0f, 1.0f}}, {0, 1, 2, 0, 2, 3}, m_FlatMesh.m_vertexSize);
        m_FlatMesh.m_attribs = {m_Shader.m_attributes.front()};
        m_FlatMesh.uploadMesh();

        m_WarpMesh.m_attribs = m_Shader.m_attributes;
        m_WarpMesh.m_isStriped = true;
    }

    return m_Target.m_fbo;
}

bool HmiPostProcessor::setColorProfile(const unsigned char *colorProfile, size_t size)
{
    if (isColorProfileValid(colorProfile, size))
    {
        m_lutTexture.setColorData(colorProfile);
        return true;
    }

    return false;
}

bool HmiPostProcessor::setWarpingTables(const unsigned char *data, size_t dataSize,
                                        size_t countOfColumns, size_t countOfRows,
                                        int32_t originWidth, int32_t originHeight,
                                        int32_t activeX, int32_t activeY,
                                        int32_t activeWidth, int32_t activeHeight)
{
    size_t cellSize = WARPING_TABLE_CELL_SIZE;
    size_t tableSize = countOfColumns * countOfRows * cellSize;

    if (data == nullptr || (data != nullptr && dataSize != 0 && dataSize % tableSize != 0))
    {
        if (HmiRender::s_callback)
        {
            std::stringstream stream;
            stream << "parseWarpingTables() wrong binary data size: " << dataSize << std::endl;
            HmiRender::s_callback(stream.str().c_str());
        }
        return false;
    }
    else
    {
        const bool isDefaultPrewarpingGeometry = (activeX == -1 && activeY == -1 &&
                                                   activeWidth == -1 && activeHeight == -1);
        size_t countOfTableItems = countOfColumns * countOfRows;
        size_t idx = 0;

        float wshift = static_cast<float>(activeX) / originWidth;
        float hshift = static_cast<float>(activeY) / originHeight;

        float wscale = static_cast<float>(activeWidth) / originWidth;
        float hscale = static_cast<float>(activeHeight) / originHeight;

        float wstep = 1.0f / (countOfColumns - 1) * (isDefaultPrewarpingGeometry ? 1 : wscale);
        float hstep = 1.0f / (countOfRows - 1) * (isDefaultPrewarpingGeometry ? 1 : hscale);

        std::vector<float> attr0, attr1;
        attr0.resize(countOfTableItems * VECTOR4);
        attr1.resize(countOfTableItems * VECTOR4);

        size_t shift = 0;
        for (size_t i = 0; i < 3; ++i)
        {
            auto &table = ((i == 0) ? attr0 : attr1);

            size_t offset = shift;
            for (size_t y = 0; y < countOfRows; ++y)
            {
                for (size_t x = 0; x < countOfColumns; ++x)
                {
                    int32_t decX = static_cast<int>( 0 << 24 | 0 << 16 | data[idx + 1] << 8 | data[idx]);
                    float tx = static_cast<float>(decX) / 64.0f;
                    int32_t decY = static_cast<int>( 0 << 24 | 0 << 16 | data[idx + 3] << 8 | data[idx + 2]);
                    float ty = static_cast<float>(decY) / 64.0f;

                    table[offset] = tx;
                    table[offset + 1] = ty;
                    if (i == 0) {
                        table[offset + 2] = wstep * x + (isDefaultPrewarpingGeometry ? 0 : wshift);
                        table[offset + 3] = 1.0f - hstep * y - (isDefaultPrewarpingGeometry ? 0 : hshift);
                    }
                    offset += VECTOR4;
                    idx += cellSize;
                }
            }
            if (i != 0)
            {
                shift += 2;
            }
        }

        m_indexCount = ((countOfColumns * 2) * (countOfRows - 1) + (2 * (countOfRows - 2)) + 1);

        std::vector<uint16_t> indices;
        indices.resize(m_indexCount);

        size_t offset = 0;
        for (size_t y = 0; y < countOfRows - 1; ++y)
        {
            if (y > 0)
            {
                indices[offset++] = static_cast<uint16_t>(y * countOfColumns);
            }
            for (uint8_t x = 0; x < countOfColumns; x++)
            {
                indices[offset++] = static_cast<uint16_t>((y * countOfColumns) + x);
                indices[offset++] = static_cast<uint16_t>((y + 1) * countOfColumns + x);
            }
            if (y < countOfColumns)
            {
                indices[offset++] = static_cast<uint16_t>(((y + 1) * countOfColumns) + (countOfColumns - 1));
            }
        }
        m_WarpMesh.setData({attr0, attr1}, indices, 0);
    }

    return true;
}

void HmiPostProcessor::turnWarpingOn(bool value)
{
    m_isWarpingOn = value;
}

void HmiPostProcessor::setWarpingPercentage(float percentage)
{
    m_warpingPercentage = percentage;
}

void HmiPostProcessor::setMask(const HmiTexture *mask, int32_t posX, int32_t posY, int32_t sizeX, int32_t sizeY)
{
    if (mask)
    {
        m_Shader.setTexture(m_maskTexture, mask->m_texture);
        m_maskPosition[0] = static_cast<float>(m_Width) / sizeX;
        m_maskPosition[1] = static_cast<float>(m_Height) / sizeY;
        m_maskPosition[2] = static_cast<float>(posX) / -static_cast<float>(sizeX);
        m_maskPosition[3] = static_cast<float>(m_Height - sizeY - posY) / -static_cast<float>(sizeY);
    }
    else
    {
        m_Shader.setTexture(m_maskTexture, m_whiteTexture.m_texture);
        m_maskOpacity[0] = 0.0f;
        m_maskOpacity[1] = 0.0f;
        m_maskOpacity[2] = 0.0f;
        m_maskOpacity[2] = 0.0f;
    }
}

void HmiPostProcessor::setMaskMode(bool map, bool guidance, bool flip)
{
    m_maskOpacity[1] = (map) ? 1.0f : 0.0f;
    m_maskOpacity[2] = (guidance) ? 1.0f : 0.0f;
    m_maskOpacity[3] = (flip) ? 1.0f : 0.0f;
}

void HmiPostProcessor::setMaskOpacity(float opacity)
{
    m_maskOpacity[0] = opacity;
}

void HmiPostProcessor::setParallax(int16_t x, int16_t y)
{
    m_OffsetX = x;
    m_OffsetY = y;
}

void HmiPostProcessor::postProcess()
{
    HmiRender::setViewport(m_OffsetX, m_OffsetY, m_Width, m_Height);
    m_WarpMesh.uploadMesh();
    m_lutTexture.uploadTexture();

    if (m_isWarpingOn)
    {
        m_mvp.ortho(0.0f, m_Width, m_Height, 0.0f, -1.0f, 1.0f);
    }
    else
    {
        m_mvp.ortho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
    }

    HmiRender::setRenderTarget(m_defaultFbo);
    HmiRender::clear();
    HmiRender::drawMesh((m_isWarpingOn) ? m_WarpMesh : m_FlatMesh, m_Shader);
    HmiRender::setViewport(0, 0, m_Width, m_Height);
}

bool HmiPostProcessor::isColorProfileValid(const unsigned char *colorProfile, size_t size)
{
    if (colorProfile == nullptr || size != (LUT_SIZE * LUT_SIZE * 4))
    {
        return false;
    }
    return true;
}
