/*
** Copyright (C) 2019 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/

#ifndef HMIPOSTPROCESSOR_H
#define HMIPOSTPROCESSOR_H

#define DEFAULT_WARPING_TABLE_COLUMNS 21
#define DEFAULT_WARPING_TABLE_ROWS 9

#define LUT_SIZE 16

#include <cstdint>
#include <cstddef>

#include "HmiRender.h"

#include "matrix4.h"

class HmiPostProcessor
{

public:
    HmiPostProcessor();

    HmiPostProcessor(const HmiPostProcessor &) = delete;
    HmiPostProcessor(HmiPostProcessor &&) = delete;
    HmiPostProcessor& operator=(const HmiPostProcessor &) = delete;
    HmiPostProcessor& operator=(HmiPostProcessor &&) = delete;

    /**
     * @brief Saves display sizes and prepares shaders with default parameters value
     * @param width - display width
     * @param height - display height
     * @param defaultFbo - default framebuffer object (in most cases equals 0)
     * @return frame buffer object
     */
    uint32_t init(uint32_t width, uint32_t height, uint32_t defaultFbo = 0);

    /**
     * @brief Updates shader with new color profile
     * @param colorProfile - the pointer to color profile data,
     * this method owns this data and must call delete [] if this data is no longer needed
     * @param size - the size of data
     * @return false if unrecognized format is received, otherwise true
     */
    bool setColorProfile(const unsigned char *colorProfile, size_t size);

    /**
     * @brief Updates shader with new set of warping tables
     * @param data - the pointer to the binary warping tables data
     * @param sizeOfData - the size of data
     * @param countOfColumns - the count of columns in warping table
     * @param countOfRows - the count of rows in warping table
     * @param width - an origin width of the window
     * @param height - an origin height of the window
     * @return false if inconsistent data is received, otherwise true
     */
    bool setWarpingTables(const unsigned char *data, size_t sizeOfData,
                          size_t countOfColumns = DEFAULT_WARPING_TABLE_COLUMNS,
                          size_t countOfRows = DEFAULT_WARPING_TABLE_ROWS,
                          int32_t originWidth = -1, int32_t originHeight = -1,
                          int32_t activeX = -1, int32_t activeY = -1,
                          int32_t activeWidth = -1, int32_t activeHeight = -1);

    /**
     * @brief Turns warping on/off
     * @param value - the on/off parameter
     */
    void turnWarpingOn(bool value);

    /**
     * @brief Recounts current warping table to a new percentage
     * @param percentage - a new percentage value
     */
    void setWarpingPercentage(float percentage);

    /**
     * @brief Enables mask to make a holes.
     * @param mask - A mask wich will be applied
     * @param posX - Position by X of the mask
     * @param posY - Position by Y of the mask
     * @param sizeX - Size by X of the mask
     * @param sizeY - Size by Y of the mask
     */
    void setMask(const HmiTexture *mask, int32_t posX, int32_t posY, int32_t sizeX, int32_t sizeY);

    /**
     * @brief Sets mask mode.
     * @param map - A map mask visiblity flag
     * @param guidance - A guidance mask visiblity flag
     * @param flip - A guidance mask wil be flipped by Y flag
     */
    void setMaskMode(bool map, bool guidance, bool flip);

    /**
     * @brief sets the mask opacity.
     * @param opacity - Opacity of the mask
     */
    void setMaskOpacity(float opacity);

    /**
     * @brief Updates parallax offset
     * @param x - a new offset by X
     * @param y - a new offset by Y
     */
    void setParallax(int16_t x, int16_t y);

    /**
     * @brief Does all necessary post processing
     */
    void postProcess();

private:
    bool isColorProfileValid(const unsigned char *colorProfile, size_t size);

    bool m_isWarpingOn;
    size_t m_warpingTableColumnsCount;
    size_t m_warpingTableRowsCount;
    float m_warpingPercentage;

    uint32_t m_Width;
    uint32_t m_Height;

    int16_t m_OffsetX;
    int16_t m_OffsetY;

    Matrix4 m_mvp;
    const unsigned char *m_colorProfile;

    float m_maskPosition[4];
    float m_maskOpacity[4];
    uint32_t m_maskTexture;

    uint32_t m_defaultFbo;
    size_t m_indexCount;

    HmiRenderTarget m_Target;
    HmiShader m_Shader;
    HmiMesh m_FlatMesh;
    HmiMesh m_WarpMesh;
    HmiTexture m_lutTexture;
    HmiTexture m_whiteTexture;
};

#endif // HMIPOSTPROCESSOR_H
