/*
** Copyright (C) 2019-2020 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/
#include "HmiRender.h"

#include <sstream>

#if defined(INTEGRITY) || defined(GENIVI)
    #include <GLES3/gl3.h>
#else
    #include "glad.h"
#endif

ErrorFun HmiRender::s_callback = nullptr;

#define ATTR0_ATTRIB 0
#define ATTR1_ATTRIB 1

#define INSTANCE_ATTRIB 2


HmiTexture::HmiTexture() :
    m_texture(0),
    m_width(0),
    m_height(0),
    m_internal(GL_RGBA8),
    m_format(GL_RGBA),
    m_isDirty(false),
    m_colorData(nullptr)
{
}

HmiTexture::~HmiTexture()
{
    if (m_texture)
    {
        glDeleteTextures(1, &m_texture);
    }
}

void HmiTexture::create(uint32_t width, uint32_t height, uint8_t *bits, Format format, bool isRepeat, bool isLinear)
{
    m_width = width;
    m_height = height;
    glGenTextures(1, &m_texture);
    glBindTexture(GL_TEXTURE_2D, m_texture);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (isLinear) ? GL_LINEAR : GL_NEAREST);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (isRepeat) ? GL_REPEAT : GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (isRepeat) ? GL_REPEAT : GL_CLAMP_TO_EDGE);

    m_internal = GL_RGBA8;
    m_format = GL_RGBA;
    switch(format)
    {
    case R:
    {
        m_internal = GL_R8;
        m_format = GL_RED;
        break;
    }
    case DEPTH:
    {
        m_internal = GL_DEPTH_COMPONENT16;
        m_format = GL_DEPTH_COMPONENT;
        break;
    }
    default: break;
    }

    glTexImage2D(GL_TEXTURE_2D, 0, m_internal,
                 static_cast<GLsizei>(m_width), static_cast<GLsizei>(m_height),
                 0, m_format, GL_UNSIGNED_BYTE, bits);
}

void HmiTexture::uploadTexture()
{
    if (m_isDirty)
    {
        glBindTexture(GL_TEXTURE_2D, m_texture);
        glTexImage2D(GL_TEXTURE_2D, 0, m_internal,
                     static_cast<GLsizei>(m_width), static_cast<GLsizei>(m_height),
                     0, m_format, GL_UNSIGNED_BYTE, m_colorData);

        m_isDirty = false;
    }
}

void HmiTexture::setColorData(const unsigned char *colorData)
{
    if (colorData)
    {
        m_colorData = colorData;
        m_isDirty = true;
    }
}

HmiRenderTarget::HmiRenderTarget() :
    m_fbo(0)
{

}

HmiRenderTarget::~HmiRenderTarget()
{
    if (m_fbo)
    {
        glDeleteFramebuffers(1, &m_fbo);
    }
}

void HmiRenderTarget::create(uint32_t width, uint32_t height, bool depth)
{
    m_color.create(width, height, nullptr, HmiTexture::RGBA);

    if (depth)
    {
        m_depth.create(width, height, nullptr, HmiTexture::DEPTH);
    }

    glGenFramebuffers(1, &m_fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, m_fbo);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_color.m_texture, 0);
    if (depth)
    {
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, m_depth.m_texture, 0);
    }
}

HmiShader::HmiShader() :
    m_isBlended(false),
    m_program(0),
    m_mvpLocation(0),
    m_mvp(nullptr)
{

}

HmiShader::~HmiShader()
{
    if (m_program)
    {
       glDeleteProgram(m_program);
    }
}

bool HmiShader::checkShader(uint32_t shader, bool link)
{
    int value = 0;

    if (!link)
    {
        glGetShaderiv(shader, GL_COMPILE_STATUS, &value);
    }
    else
    {
        glGetProgramiv(shader, GL_LINK_STATUS, &value);
    }

    if (value != GL_TRUE)
    {
        if (!link)
        {
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &value);
        }
        else
        {
            glGetProgramiv(shader, GL_INFO_LOG_LENGTH, &value);
        }
        if (value)
        {
            char *buff  = new char[static_cast<size_t>(value + 1)];
            if (!link)
            {
                glGetShaderInfoLog(shader, value, nullptr, buff);
            }
            else
            {
                glGetProgramInfoLog(shader, value, nullptr, buff);
            }

            if (HmiRender::s_callback)
            {
                std::stringstream stream;
                stream << "Building program error: " << buff << std::endl;
                HmiRender::s_callback(stream.str().c_str());
            }
            delete []buff;
        }
        return false;
    }
    return true;
}

uint32_t HmiShader::buildShader(uint32_t type, const char *data)
{
    uint32_t shader = glCreateShader(type);
    if (shader)
    {
        glShaderSource(shader, 1, &data, nullptr);
        glCompileShader(shader);
        checkShader(shader, false);
    }
    return shader;
}

void HmiShader::bind() const
{
    glUseProgram(m_program);

    for(auto it : m_uniforms)
    {
        switch(it.second.first)
        {
        case 1:
        {
            glUniform1fv(it.first, 1, it.second.second);
            break;
        }
        case 2:
        {
            glUniform2fv(it.first, 1, it.second.second);
            break;
        }
        case 3:
        {
            glUniform3fv(it.first, 1, it.second.second);
            break;
        }
        case 4:
        {
            glUniform4fv(it.first, 1, it.second.second);
            break;
        }
        default: break;
        }
    }

    for(uint32_t i = 0; i < m_textures.size(); i++)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_2D, m_textures[i]);
    }
    glUniformMatrix4fv(m_mvpLocation, 1, GL_FALSE, m_mvp);

    if(m_isBlended)
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glBlendEquation(GL_FUNC_ADD);
    }
}

void HmiShader::unbind() const
{
    for(uint32_t i = 0; i < m_textures.size(); i++)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_2D, 0);
    }

    glActiveTexture(GL_TEXTURE0);
    glUseProgram(0);
}

bool HmiShader::load(const char *vertexShader, const char *fragmentShader)
{
    uint32_t vertex = buildShader(GL_VERTEX_SHADER, vertexShader);
    uint32_t fragment = buildShader(GL_FRAGMENT_SHADER, fragmentShader);

    if (vertex && fragment)
    {
        m_program = glCreateProgram();
        if (m_program)
        {
            glAttachShader(m_program, vertex);
            glAttachShader(m_program, fragment);
            glLinkProgram(m_program);
            glDetachShader(m_program, vertex);
            glDetachShader(m_program, fragment);

            checkShader(m_program, true);

            glDeleteShader(vertex);
            glDeleteShader(fragment);

            glUseProgram(m_program);

            m_mvpLocation = glGetUniformLocation(m_program, "mvp");

            return true;
        }
    }
    return false;
}

void HmiShader::setAttributes(const std::list<const char *> &attribs)
{
    for (auto it : attribs)
    {
        m_attributes.push_back(glGetAttribLocation(m_program, it));
    }
}

uint32_t HmiShader::addTexture(const char *name, uint32_t id)
{
    uint32_t result = m_textures.size();
    glUniform1i(glGetUniformLocation(m_program, name), result);
    m_textures.push_back(id);
    return result;
}

void HmiShader::setTexture(uint32_t pos, uint32_t id)
{
    m_textures[pos] = id;
}

void HmiShader::addUniform(const char *name, float *ptr, uint32_t size)
{
    m_uniforms[glGetUniformLocation(m_program, name)] = (std::make_pair(size, ptr));
}

HmiMesh::HmiMesh() :
    m_isDirty(false),
    m_isInstanced(true),
    m_isStriped(false),
    m_indexBuff(0),
    m_vertexBuff(nullptr),
    m_vao(0),
    m_instance(0),
    m_vertexSize(VECTOR4)
{

}

HmiMesh::~HmiMesh()
{
    glDeleteBuffers(1, &m_indexBuff);
    glDeleteBuffers(m_vertices.size(), m_vertexBuff);

    delete []m_vertexBuff;
}

void HmiMesh::setData(const std::list<std::vector<float>> &vertices, const std::vector<uint16_t> &indices, uint32_t vertexSize)
{
    m_vertices = vertices;
    m_indices = indices;
    m_vertexSize = vertexSize;

    m_isDirty = true;

    if (m_vertexBuff != nullptr)
    {
        delete []m_vertexBuff;
    }

    m_vertexBuff = new uint32_t[m_vertices.size()];
}

void HmiMesh::setAttribs(const std::list<uint32_t> &attribs)
{
    m_attribs = attribs;
}

void HmiMesh::uploadMesh()
{
    if (m_isDirty)
    {
        if (m_indexBuff == 0)
        {
            glGenBuffers(1, &m_indexBuff);
            glGenBuffers(m_vertices.size(), m_vertexBuff);
        }

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_indexBuff);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, static_cast<GLsizeiptr>(sizeof(uint16_t) * m_indices.size()), &m_indices[0], GL_DYNAMIC_DRAW);

        int32_t v = 0;
        for (auto it : m_vertices)
        {
            glBindBuffer(GL_ARRAY_BUFFER, m_vertexBuff[v]);
            glBufferData(GL_ARRAY_BUFFER, static_cast<GLsizeiptr>(it.size() * sizeof(float)), &it[0], GL_DYNAMIC_DRAW);
            v++;
        }

        if (m_isInstanced && m_instance == 0)
        {
            glGenBuffers(1, &m_instance);
        }

        if (m_vao == 0)
        {
            glGenVertexArrays(1, &m_vao);
            glBindVertexArray(m_vao);

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_indexBuff);

            v = 0;
            size_t offset = 0;
            for (auto it : m_attribs)
            {
                glBindBuffer(GL_ARRAY_BUFFER, m_vertexBuff[v]);
                glEnableVertexAttribArray(it);
                glVertexAttribPointer(it, 4, GL_FLOAT, GL_FALSE, static_cast<GLsizei>(m_vertexSize * sizeof(float)), reinterpret_cast<void *>(offset));

                offset += m_vertexSize * sizeof(float);
                v++;
            }

            if (m_isInstanced)
            {
                glBindBuffer(GL_ARRAY_BUFFER, m_instance);
                for (uint32_t i = 0; i < 4; i++)
                {
                    glEnableVertexAttribArray(INSTANCE_ATTRIB + i);
                    glVertexAttribPointer(INSTANCE_ATTRIB + i, 4, GL_FLOAT, GL_FALSE, MATRIX4_SIZE, reinterpret_cast<void *>(i * VECTOR4_SIZE));
                    glVertexAttribDivisor(INSTANCE_ATTRIB + i, 1);
                }
            }

            glBindVertexArray(0);
        }

        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

        m_isDirty = false;
    }
}

void HmiRender::setErrorCallback(ErrorFun fun)
{
    s_callback = fun;
}

int32_t HmiRender::init()
{
#if !defined (INTEGRITY) && !defined (GENIVI)
    return gladLoadGL();
#else
    return 1;
#endif
}

void HmiRender::setViewport(int x, int y, int width, int height)
{
    glViewport(x, y, width, height);
}

void HmiRender::clear()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void HmiRender::clearDepth()
{
    glDepthMask(GL_TRUE);
    glClearDepthf(1.0f);
    glClear(GL_DEPTH_BUFFER_BIT);
}

void HmiRender::drawMesh(const HmiMesh &mesh, const HmiShader &shader)
{
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    shader.bind();

    glBindVertexArray(mesh.m_vao);

    glDrawElements((mesh.m_isStriped) ? GL_TRIANGLE_STRIP : GL_TRIANGLES, static_cast<GLsizei>(mesh.m_indices.size()), GL_UNSIGNED_SHORT, nullptr);

    glBindVertexArray(0);

    shader.unbind();
}

void HmiRender::drawMeshInstanced(const HmiMesh &mesh, const HmiShader &shader, const float *params, uint32_t count)
{
    shader.bind();

    glBindBuffer(GL_ARRAY_BUFFER, mesh.m_instance);
    glBufferData(GL_ARRAY_BUFFER, count * MATRIX4_SIZE, params, GL_DYNAMIC_DRAW);

    glBindVertexArray(mesh.m_vao);

    glDrawElementsInstanced((mesh.m_isStriped) ? GL_TRIANGLE_STRIP : GL_TRIANGLES, static_cast<GLsizei>(mesh.m_indices.size()), GL_UNSIGNED_SHORT, nullptr, count);

    glBindVertexArray(0);

    shader.unbind();
}

void HmiRender::setRenderTarget(uint32_t target)
{
    glBindFramebuffer(GL_FRAMEBUFFER, target);
}
