/*
** Copyright (C) 2019-2020 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/
#ifndef HMIRENDER_H
#define HMIRENDER_H

#include <list>
#include <vector>
#include <unordered_map>

typedef void (* ErrorFun)(const char *);

#define MATRIX4 16
#define VECTOR4 4

#define MATRIX4_SIZE MATRIX4 * sizeof(float)
#define VECTOR4_SIZE VECTOR4 * sizeof(float)

#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))

struct HmiTexture
{
    enum Format
    {
        R,
        RGBA,
        DEPTH
    };

    HmiTexture();
    ~HmiTexture();

    void create(uint32_t width, uint32_t height, uint8_t *bits, Format format = RGBA, bool isRepeat = false, bool isLinear = true);
    void uploadTexture();
    void setColorData(const unsigned char *colorData);

    uint32_t m_texture;
    uint32_t m_width;
    uint32_t m_height;

    uint32_t m_internal;
    uint32_t m_format;

    bool m_isDirty;
    const unsigned char *m_colorData;
};

struct HmiRenderTarget
{
    HmiRenderTarget();
    ~HmiRenderTarget();

    void create(uint32_t width, uint32_t height, bool depth = false);

    uint32_t m_fbo;

    HmiTexture m_color;
    HmiTexture m_depth;
};

struct HmiShader
{
    HmiShader();
    ~HmiShader();

    bool load(const char *vertex, const char *fragment);
    void setAttributes(const std::list<const char *> &attribs);
    uint32_t addTexture(const char *name, uint32_t id);
    void setTexture(uint32_t pos, uint32_t id);
    void addUniform(const char *name, float *ptr, uint32_t size);
    bool checkShader(uint32_t shader, bool link);
    uint32_t buildShader(uint32_t type, const char *data);
    void bind() const;
    void unbind() const;

    bool m_isBlended;
    uint32_t m_program;
    uint32_t m_mvpLocation;
    float *m_mvp;
    std::vector<uint32_t> m_textures;
    std::unordered_map<uint32_t, std::pair<uint32_t, float *>> m_uniforms;
    std::list<uint32_t> m_attributes;
};

struct HmiMesh
{
    HmiMesh();
    ~HmiMesh();

    void setData(const std::list<std::vector<float>> &vertices, const std::vector<uint16_t> &indices, uint32_t vertexSize);
    void setAttribs(const std::list<uint32_t> &attribs);
    void uploadMesh();

    bool m_isDirty;
    bool m_isInstanced;
    bool m_isStriped;
    uint32_t m_indexBuff;
    uint32_t *m_vertexBuff;
    uint32_t m_vao;
    uint32_t m_instance;
    uint32_t m_vertexSize;
    std::list<std::vector<float>> m_vertices;
    std::vector<uint16_t> m_indices;
    std::list<uint32_t> m_attribs;
};

class HmiRender
{
public:
    static int32_t init();

    static void clear();
    static void clearDepth();

    static void drawMesh(const HmiMesh &mesh, const HmiShader &shader);

    static void drawMeshInstanced(const HmiMesh &mesh, const HmiShader &shader, const float *params, uint32_t count);

    static void setRenderTarget(uint32_t target);

    /**
     * @brief Registers an error collback function
     * @param fun - callback function
     */
    static void setErrorCallback(ErrorFun fun);

    static ErrorFun s_callback;

    static void setViewport(int x, int y, int width, int height);
};

#endif // HMIRENDER_H
