/*
** Copyright (C) 2019 HARMAN Connected Services
** All rights reserved
**
** This software embodies materials and concepts which are
** confidential to HARMAN Connected Services and is made available solely
** pursuant to the terms of a written license agreement with
** HARMAN Connected Services.
**
** Designed and Developed by HARMAN Connected Services
**
*/
#include "matrix4.h"
#include <math.h>

#define PI 3.14159265358979323846f
#define DEG2RAD (PI / 180.0f)

Matrix4::Matrix4()
{
    identity();
}

bool Matrix4::operator==(const Matrix4 &matrix) const
{
    for(int i = 0; i < 16; i++)
    {
        if(mat[i] != matrix.mat[i])
        {
            return false;
        }
    }
    return true;
}

bool Matrix4::operator!=(const Matrix4 &matrix) const
{
    return !(*this == matrix);
}

float &Matrix4::operator[](int i)
{
    return mat[i];
}

float Matrix4::operator[](int i) const
{
    return mat[i];
}

Matrix4 Matrix4::operator*(float factor) const
{
    Matrix4 ret;
    ret[0] = mat[0] * factor; ret[4] = mat[4] * factor; ret[8]  = mat[8]  * factor; ret[12] = mat[12] * factor;
    ret[1] = mat[1] * factor; ret[5] = mat[5] * factor; ret[9]  = mat[9]  * factor; ret[13] = mat[13] * factor;
    ret[2] = mat[2] * factor; ret[6] = mat[6] * factor; ret[10] = mat[10] * factor; ret[14] = mat[14] * factor;
    ret[3] = mat[3] * factor; ret[7] = mat[7] * factor; ret[11] = mat[11] * factor; ret[15] = mat[15] * factor;
    return ret;
}

Matrix4 Matrix4::operator*(const Matrix4 &matrix) const
{
    Matrix4 ret;
    ret[0]  = mat[0] * matrix[0]  + mat[4] * matrix[1]  + mat[8]  * matrix[2]  + mat[12] * matrix[3];
    ret[1]  = mat[1] * matrix[0]  + mat[5] * matrix[1]  + mat[9]  * matrix[2]  + mat[13] * matrix[3];
    ret[2]  = mat[2] * matrix[0]  + mat[6] * matrix[1]  + mat[10] * matrix[2]  + mat[14] * matrix[3];
    ret[3]  = mat[3] * matrix[0]  + mat[7] * matrix[1]  + mat[11] * matrix[2]  + mat[15] * matrix[3];
    ret[4]  = mat[0] * matrix[4]  + mat[4] * matrix[5]  + mat[8]  * matrix[6]  + mat[12] * matrix[7];
    ret[5]  = mat[1] * matrix[4]  + mat[5] * matrix[5]  + mat[9]  * matrix[6]  + mat[13] * matrix[7];
    ret[6]  = mat[2] * matrix[4]  + mat[6] * matrix[5]  + mat[10] * matrix[6]  + mat[14] * matrix[7];
    ret[7]  = mat[3] * matrix[4]  + mat[7] * matrix[5]  + mat[11] * matrix[6]  + mat[15] * matrix[7];
    ret[8]  = mat[0] * matrix[8]  + mat[4] * matrix[9]  + mat[8]  * matrix[10] + mat[12] * matrix[11];
    ret[9]  = mat[1] * matrix[8]  + mat[5] * matrix[9]  + mat[9]  * matrix[10] + mat[13] * matrix[11];
    ret[10] = mat[2] * matrix[8]  + mat[6] * matrix[9]  + mat[10] * matrix[10] + mat[14] * matrix[11];
    ret[11] = mat[3] * matrix[8]  + mat[7] * matrix[9]  + mat[11] * matrix[10] + mat[15] * matrix[11];
    ret[12] = mat[0] * matrix[12] + mat[4] * matrix[13] + mat[8]  * matrix[14] + mat[12] * matrix[15];
    ret[13] = mat[1] * matrix[12] + mat[5] * matrix[13] + mat[9]  * matrix[14] + mat[13] * matrix[15];
    ret[14] = mat[2] * matrix[12] + mat[6] * matrix[13] + mat[10] * matrix[14] + mat[14] * matrix[15];
    ret[15] = mat[3] * matrix[12] + mat[7] * matrix[13] + mat[11] * matrix[14] + mat[15] * matrix[15];
    return ret;
}

Matrix4 Matrix4::operator+(const Matrix4 &matrix) const
{
    Matrix4 ret;
    ret[0] = mat[0] + matrix[0]; ret[4] = mat[4] + matrix[4]; ret[8]  = mat[ 8] + matrix[ 8]; ret[12] = mat[12] + matrix[12];
    ret[1] = mat[1] + matrix[1]; ret[5] = mat[5] + matrix[5]; ret[9]  = mat[ 9] + matrix[ 9]; ret[13] = mat[13] + matrix[13];
    ret[2] = mat[2] + matrix[2]; ret[6] = mat[6] + matrix[6]; ret[10] = mat[10] + matrix[10]; ret[14] = mat[14] + matrix[14];
    ret[3] = mat[3] + matrix[3]; ret[7] = mat[7] + matrix[7]; ret[11] = mat[11] + matrix[11]; ret[15] = mat[15] + matrix[15];
    return ret;
}

Matrix4 Matrix4::operator-(const Matrix4 &matrix) const
{
    Matrix4 ret;
    ret[0] = mat[0] - matrix[0]; ret[4] = mat[4] - matrix[4]; ret[8]  = mat[ 8] - matrix[ 8]; ret[12] = mat[12] - matrix[12];
    ret[1] = mat[1] - matrix[1]; ret[5] = mat[5] - matrix[5]; ret[9]  = mat[ 9] - matrix[ 9]; ret[13] = mat[13] - matrix[13];
    ret[2] = mat[2] - matrix[2]; ret[6] = mat[6] - matrix[6]; ret[10] = mat[10] - matrix[10]; ret[14] = mat[14] - matrix[14];
    ret[3] = mat[3] - matrix[3]; ret[7] = mat[7] - matrix[7]; ret[11] = mat[11] - matrix[11]; ret[15] = mat[15] - matrix[15];
    return ret;
}

Matrix4 &Matrix4::operator*=(float factor)
{
    return *this = *this * factor;
}

Matrix4 &Matrix4::operator*=(const Matrix4 &matrix)
{
    return *this = *this * matrix;
}

Matrix4 &Matrix4::operator+=(const Matrix4 &matrix)
{
    return *this = *this + matrix;
}

Matrix4 &Matrix4::operator-=(const Matrix4 &matrix)
{
    return *this = *this - matrix;
}

void Matrix4::zero()
{
    mat[0] = 0.0f; mat[4] = 0.0f; mat[ 8] = 0.0f; mat[12] = 0.0f;
    mat[1] = 0.0f; mat[5] = 0.0f; mat[ 9] = 0.0f; mat[13] = 0.0f;
    mat[2] = 0.0f; mat[6] = 0.0f; mat[10] = 0.0f; mat[14] = 0.0f;
    mat[3] = 0.0f; mat[7] = 0.0f; mat[11] = 0.0f; mat[15] = 0.0f;
}

void Matrix4::identity()
{
    mat[0] = 1.0f; mat[4] = 0.0f; mat[ 8] = 0.0f; mat[12] = 0.0f;
    mat[1] = 0.0f; mat[5] = 1.0f; mat[ 9] = 0.0f; mat[13] = 0.0f;
    mat[2] = 0.0f; mat[6] = 0.0f; mat[10] = 1.0f; mat[14] = 0.0f;
    mat[3] = 0.0f; mat[7] = 0.0f; mat[11] = 0.0f; mat[15] = 1.0f;
}

void Matrix4::rotate(float angle, float x, float y, float z)
{
    float rad = angle * DEG2RAD;
    float c = (rad == PI * 0.5f) ? 0 : cosf(rad);
    float s = (rad == PI) ? 0 : sinf(rad);

    float xy = x * y;
    float yz = y * z;
    float zx = z * x;
    float xs = x * s;
    float ys = y * s;
    float zs = z * s;
    mat[0] = (1.0f - c) * x * x + c; mat[4] = (1.0f - c) * xy - zs;   mat[ 8] = (1.0f - c) * zx + ys;
    mat[1] = (1.0f - c) * xy + zs;   mat[5] = (1.0f - c) * y * y + c; mat[ 9] = (1.0f - c) * yz - xs;
    mat[2] = (1.0f - c) * zx - ys;   mat[6] = (1.0f - c) * yz + xs;   mat[10] = (1.0f - c) * z * z + c;
}

void Matrix4::scale(float x, float y, float z)
{
    mat[0] = x;    mat[4] = 0.0f; mat[ 8] = 0.0f; mat[12] = 0.0f;
    mat[1] = 0.0f; mat[5] = y;    mat[ 9] = 0.0f; mat[13] = 0.0f;
    mat[2] = 0.0f; mat[6] = 0.0f; mat[10] = z;    mat[14] = 0.0f;
    mat[3] = 0.0f; mat[7] = 0.0f; mat[11] = 0.0f; mat[15] = 1.0f;
}

void Matrix4::translate(float x, float y, float z)
{
    mat[0] = 1.0f; mat[4] = 0.0f; mat[ 8] = 0.0f; mat[12] = x;
    mat[1] = 0.0f; mat[5] = 1.0f; mat[ 9] = 0.0f; mat[13] = y;
    mat[2] = 0.0f; mat[6] = 0.0f; mat[10] = 1.0f; mat[14] = z;
    mat[3] = 0.0f; mat[7] = 0.0f; mat[11] = 0.0f; mat[15] = 1.0f;
}

void Matrix4::ortho(float left, float right, float bottom, float top, float znear, float zfar)
{
    mat[0]  =  2.0f / (right - left);
    mat[1]  =  0.0f;
    mat[2]  =  0.0f;
    mat[3]  =  0.0f;
    mat[4]  =  0.0f;
    mat[5]  =  2.0f / (top - bottom);
    mat[6]  =  0.0f;
    mat[7]  =  0.0f;
    mat[8]  =  0.0f;
    mat[9]  =  0.0f;
    mat[10] = -2.0f / (zfar - znear);
    mat[11] =  0.0f;
    mat[12] = -((right + left) / (right - left));
    mat[13] = -((top + bottom) / (top - bottom));
    mat[14] = -((zfar + znear) / (zfar - znear));
    mat[15] =  1.0f;
}
